<?php

/**
 * @file
 * Theme preprocessor functions.
 */
use Drupal\user\Entity\User;
use Drupal\Core\Render\Markup;
use Drupal\Core\Render\Element;
use Drupal\mcapi\Functions;

/**
 * Theme preprocessor for 'wallet_wrapper'.
 */
function template_preprocess_mc_wallet(&$vars) {
  $vars['canonical'] = $vars['element']['#mc_wallet']->toUrl('canonical')->toString();
  $vars['attributes']['class'] = ['wallet', $vars['element']['#view_mode']];
}

/**
 * Theme preprocessor for 'wallet_component'.
 */
function template_preprocess_mc_wallet_component(&$vars) {
  $vars['title'] = @$vars['element']['#title'];
  $vars['attributes']['class'][] = 'component';
  if (isset($vars['element']['#theme'])) {
    $vars['attributes']['class'][] = str_replace('_', '-', substr($vars['element']['#theme'], 7));
  }
}

/**
 * Theme preprocessor for all of an entity's wallets.
 */
function template_preprocess_mc_wallets(&$vars) {
  $viewbuilder = \Drupal::entityTypeManager()->getViewBuilder('mc_wallet');
  // Convert each wallet into a render array.
  foreach ($vars['wallets'] as $key => $wallet) {
    $vars['wallets'][$key] = $viewbuilder->view($wallet, $vars['view_mode']);
  }
}

/**
 * Theme preprocessor for 'balance_bars'.
 *
 * All values sent to the theme system as numbers should be formatted plainly,
 * all labels formatted normally.
 */
function template_preprocess_mc_wallet_balance_bars(&$vars) {
  $formatter = \Drupal::Service('mcapi.currency_formatter');
  $wallet = $vars['wallet'];
  $id = "given-received" . $wallet->id();
  $raw_axis_markers = $formatter->axisMarkers(max($wallet->gross_in, $wallet->gross_out));
  $vars += [
    'id' => $id,
    'functionname' => str_replace('-', '_', $id),
    'incoming' => $formatter->formatAsFloat($wallet->gross_in),
    'outgoing' => $formatter->formatAsFloat($wallet->gross_out),
    'show_in'  => $formatter->format($wallet->gross_in),
    'show_out' => $formatter->format($wallet->gross_out),
    'max' => $formatter->formatAsFloat(end($raw_axis_markers)),
    'vaxislabels' => []
  ];
  foreach ($raw_axis_markers as $marker) {
    $vars['vaxislabels'][] = [
      'value' => $formatter->formatAsFloat($marker),
      'label' => $formatter->format($marker)
    ];
  }
}

/**
 * Theme preprocessor for 'balance_viz'.
 */
function template_preprocess_mc_wallet_balance_viz(&$vars) {
  $wallet = $vars['wallet'];
  if ($wallet->gross_in > $wallet->gross_out) {
    $vars['side'] = 'positive';
  }
  elseif ($wallet->gross_in < $wallet->gross_out) {
    $vars['side'] = 'negative';
  }
  else {
    $vars['side'] = 'balanced';
  }
  $vars['balance'] = $wallet->balance;
  $vars['income_formatted'] = Functions::formatCurrency($wallet->gross_in, FALSE);
  $vars['expenditure_formatted'] = Functions::formatCurrency($wallet->gross_out, FALSE);
  $vars['income_formatted'] = Functions::formatCurrency($wallet->gross_in, FALSE);
  $vars['expenditure_formatted'] = Functions::formatCurrency($wallet->gross_out, FALSE);
  $vars['balance_formatted'] = Functions::formatCurrency($wallet->balance, FALSE);
  $vars['width'] = '15'; // em.
}

/**
 * Theme preprocessor for 'mc_wallet_stats'.
 */
function template_preprocess_mc_wallet_stats(&$vars) {
  $formatter = \Drupal::Service('mcapi.currency_formatter');
  $vars['gross_in'] = $formatter->format($vars['wallet']->gross_in);
  $vars['gross_out'] = $formatter->format($vars['wallet']->gross_out);
  $vars['volume'] = $formatter->format($vars['wallet']->volume);
  $vars['balance'] = $formatter->format($vars['wallet']->balance);
  $vars['partners'] = $vars['wallet']->partners;
  $vars['trades'] = $vars['wallet']->trades;
}

/**
 * Theme preprocessor for 'mc_main_entry'.
 */
function template_preprocess_mc_main_entry(&$vars) {
  $elements = $vars['elements'];
  $vars['payee'] = $elements['#mc_entry']->payee->label();
  $vars['payer'] = $elements['#mc_entry']->payer->label();
  $vars['description'] = $elements['#mc_entry']->description;
  $quant = \Drupal::Service('mcapi.currency_formatter')->format($elements['#mc_entry']->quant);
  $vars['quant'] = Markup::create('<div class="mc-quantity">'.$quant.'</div>');
}

/**
 * Theme preprocessor for 'mc_transaction'.
 */
function template_preprocess_mc_transaction(&$vars) {
  $elements = $vars['elements'];
  $renderer = \Drupal::service('renderer');
  // Render the first and additional entries into separate variables.
  if ($other_entries = mcapi_split_entries($elements['entries'])) {
    $vars['entries_other'] = $renderer->render($other_entries);
  }

  $vars['entry_main'] = $renderer->render($elements['entries']);
  $vars['signatures'] = $renderer->render($elements['signatures']);
  $vars['created'] = $renderer->render($elements['created'][0]);
  $other_fields = [];
  foreach (Element::children($elements) as $key) { // unrendered fields
    if (!isset($vars[$key])) {
      $vars[$key] = $other_fields[$key] = $renderer->render($elements[$key]);
    }
  }
  unset($other_fields['entries']);

  $vars['other_fields'] = Markup::create(implode($other_fields));
  if (!empty($elements['#watermark'])) {
    // To add a watermark to the render array, add #watermark => 'blah'
    $vars['watermark'] = Markup::create('<div class = "watermark">'.$elements['#watermark'].'</div>');
  }
  if (isset($elements['#links'])) {
    $vars['links'] = $renderer->render($elements['#links']);
  }
}

/**
 * Separate out the main entry from the other entries.
 * @param array $render_element
 * @return array
 *   Render array for the other entries.
 */
function mcapi_split_entries(array &$render_element) : array {
  $second_element = [];
  $keys = Element::children($render_element);
  if (count($keys) > 1) {
    array_shift($keys); // get rid of the 0 key
    $second_element = $render_element;
    unset($second_element[0]);
    foreach ($keys as $key) {
      unset($render_element[$key]);
    }
  }
  return $second_element;
}

/**
 * Theme preprocessor for 'mc_wallet_balance_history'.
 *
 * Adds a column to the data with the formatted currency value.
 */
function template_preprocess_mc_wallet_balance_history(&$vars) {
  $vars['id'] = 'timeline_'.$vars['wallet']->id();
  $points = \Drupal::entityTypeManager()
    ->getStorage('mc_transaction')
    ->historyOfWallet($vars['wallet']->id());
  $points[\Drupal::time()->getRequestTime()] = end($points);

  $formatter = \Drupal::Service('mcapi.currency_formatter');
  $raw_markers = $formatter->gChartAxes(min($points), max($points));
  // Apply smoothing, or even roughing.
  if (count($points) < $vars['width'] / 3) {
    // Step method, for a small number of points.
    $times = $values = [];
    // Make two values for each one in the keys and values.
    foreach ($points as $time => $bal) {
      $times[] = $time;
      $times[] = $time + 1;
      $bal_float = $formatter->formatAsFloat($bal);
      $values[] = $bal_float;
      $values[] = $bal_float;
    }
    // Now slide the arrays against each other to create steps.
    array_pop($values);
    array_shift($times);
    $vars['points'] = array_combine($times, $values);
  }
  elseif (count($points) > $vars['width']) {
    // Decimate the array, for a large number of points.
    $ratio = count($points) / $vars['width'];
    $factor = intval($ratio + 1);
    // Now iterate through the array taking 1 out of every $factor values.
    $i = 0;
    foreach ($points as $key => &$value) {
      if ($i % $factor != 0) {
        unset($points[$key]);
      }
      else {
        $value = $formatter->formatAsFloat($value);
      }
      $i++;
    }
    $vars['points'] = $points;
  }
  // Prepare for javascript
  foreach ($vars['points'] as $timestamp => $val) {
    $vars['daterows'][date('Y,m,d', $timestamp)] = $val;
  }
  foreach ($raw_markers as $raw_marker) {
    $vars['vaxislabels'][] = [
      'value' => $formatter->formatAsFloat($raw_marker),
      'label' => $formatter->format($raw_marker)
    ];
  }
}

/**
 * Implements template_preprocess_THEMEHOOK().
 * @note this is probably renamed.
 */
function template_preprocess_mc_signatures(&$vars) {
  $vars['name'] = User::load($vars['uid'])->getDisplayName();
  if ($vars['signed']) {
    $vars['date'] = \Drupal::service('date.formatter')->format($vars['signed']);
  }
}
