<?php

namespace Drupal\mcapi_cc;

use CreditCommons\Exceptions\CCFailure;
use CreditCommons\Exceptions\CCError;
use CreditCommons\Exceptions\MaxLimitViolation;
use CreditCommons\Workflow as CcWorkflow;
use Drupal\mcapi\Entity\Workflow as McWorkflow;
use Drupal\mcapi_cc\Functions;
use Drupal\mcapi_forms\Entity\McForm;
use Drupal\Core\Config\ConfigFactory;
use Drupal\Core\Logger\LoggerChannel;
use Drupal\Core\Render\Markup;
use GuzzleHttp\RequestOptions;
use Drupal\Core\Messenger\Messenger;


/**
 * Service handling outgoing requests to the trunkward node.
 */
class TrunkwardNodeRequester extends \CreditCommons\NodeRequester {

  private $logger;
  private $config;
  private $messenger;

  /**
   * Constructor
   *
   * @param ConfigFactory $config_factory
   * @param LoggerChannel $logger_channel
   * @param Messenger $messenger
   */
  public function __construct(ConfigFactory $config_factory, LoggerChannel $logger_channel, Messenger $messenger) {
    $this->config = $config_factory->get('mcapi_cc.settings');
    Functions::setErrorContext($this->config);
    $this->logger = $logger_channel;
    $this->messenger = $messenger;
    if ($this->configured()) {
      // Initiate the service from the settings.
      parent::__construct(
        $this->config->get('trunkward.url'),
        $this->config->get('node_name'),
        ''
      );
    }
    $this->catch = TRUE;
    $this->timeout = $this->config->get('trunkward.timeout') ?: 2;
  }

  public function configured() : bool {
    return $this->config->get('node_name') and $this->config->get('trunkward.url') and $this->config->get('trunkward.id');
  }

  /**
   * Update the workflows from the trunkward node.
   *
   * @note workflows should change very rarely and even then should be
   * deprecated (disabled) long before being removed to allow transactions to complete.
   * There is currently no procedure for converting transactions with old workflows to new.
   *
   * On update separate wfs into local only, remoteonly and both
   *   foreach localonly
   *     ensure the ->isLocal() is true.
   *   foreach remoteonly
   *     if a remote name is the same as a local name
   *       if the config->isLocal() is true
   *         rename the local config with local_ prefix && rewrite the ledger replacing the type name.
   *     copy the remote to config with ->isLocal()=false
   *   foreach both
   *     ensure the transactionType has isLocal()->false
   *     if the names are different
   *       rewrite the ledger to use the remote name.
   */
  public function updateWorkflows() {
    $localised_wfs = $trunkward_wfs = $config_wfs = [];
    // List the remote and local workflows keyed by hash
    foreach ($this->getWorkflows() as $hash => $wf) {
      $trunkward_wfs[$hash] = $wf;
    }
    foreach (McWorkflow::loadMultiple() as $mcType) {
      $type = new \stdClass;
      foreach (get_object_vars($mcType) as $varname => $val) {
        $type->{$varname} = $val;
      }
      $wf = new CcWorkflow($type);
      $hash = $wf->getHash();
      if (isset($trunkward_wfs[$hash])) {
        $trunkward_wfs[$hash]->localId = $mcType->id();
      }
      $config_wfs[$hash] = $wf;
    }

    $remote_wfs = array_diff_key($trunkward_wfs, $config_wfs);
    $local_wfs = array_diff_key($config_wfs, $trunkward_wfs);
    $common_wfs = array_intersect_key($trunkward_wfs, $config_wfs);


    $clear_cache = FALSE;
    $home = \Drupal::config('mcapi_cc.settings')->get('node_name');
    foreach ($local_wfs as $local_wf) {
      /** @var \Drupal\mcapi\Entity\Type $local_wf */
      // All the items without trunkward equivalents should have the local flag set to TRUE
      $mcType = McWorkflow::load($local_wf->id);
      if ($mcType->home != $home) {
        $mcType->home = $home;
        $mcType->save();
        $clear_cache = TRUE;
        $this->logger->notice(
          'Workflow @name is no longer available remotely and is now considered a local workflow',
          ['@name' => $mcType->id]
        );
      }
      $labels[] = $local_wf->label;
    }
    foreach ($remote_wfs as $remote_wf) {
      /** @var CreditCommons\Workflow $remote_wf */
      // All trunkward workflows must be represented by local config.
      if ($mcType = McWorkflow::load($remote_wf->id)) {
        if ($mcType->home <> $home) {
          $this->logger->error("Shouldn't be possible to have a remote workflow with and a different hash but the same name as a non-local transaction type.");
        }
        Functions::localiseTransactionTypeName($mcType); // Also updates existing transactions
        $this->logger->notice(
          'Workflow @name was the same as a remote workflow and has been renamed to @new_name',
          ['@name' => $remote_wf->id, '@new_name' => $mcType->id()]
        );
      }
      mcapi_cc_convert_credcom_wf_to_mc_workflow($remote_wf)->save();
      $clear_cache = TRUE;
      $this->logger->notice(
        'The remote workflow @name has been saved locally',
        ['@name' => $remote_wf->id]
      );
        $clear_cache = TRUE;
      $labels[] = $remote_wf->label;
    }
    foreach ($common_wfs as $hash => $common_wf) {
      /** @var CreditCommons\Workflow $common_wf */
      // Common wfs must have the same name with the config->local set to false
      $mcType = McWorkflow::load($common_wf->localId);
      if ($common_wf->id <> $mcType->id) {
        Functions::updateTransactionTypes($mcType->id(), $common_wf->id);
        $map[$mcType->id()] = $common_wf->id;
        $mcType->delete();
        $clear_cache = TRUE;
        mcapi_cc_convert_credcom_wf_to_mc_workflow($common_wf)->save();
        $this->logger->notice(
          'Replaced transactionType @type with remote workflow @id which had the same hash.',
          ['@type' => $mcType->id(), '@id' => $common_wf->id]
        );
      }
      elseif ($mcType->home == $home) {
        $mcType->home = $common_wf->home;
        $mcType->save();
        $clear_cache = TRUE;
        $this->logger->notice('Transaction type @type was converted to a local version of a remote workflow.');
      }
      $labels[] = $common_wf->label;
    }
    if (count($labels) > count(array_unique($labels))) {
      $this->logger->warning('Two workflows share the same label.');
    }
    if ($clear_cache) {
      \Drupal::service('cache_tags.invalidator')->invalidateTags(['entity_type:mc_workflow']);
    }
    // untested!
    // Change the workflow names in existing mc_forms
    if (class_exists('\Drupal\mcapi_forms\Entity\McForm')) {
      foreach (McForm::loadMultiple() as $form) {
        if (isset($map[$form->getWorkflowId()])) {
          $form->setWorkflowId($map[$form->getWorkflowId()])->save();
        }
      }
    }
  }

  /**
   * {@inheritDoc}
   *
   * Theoretically this could be disintermediated by javascript calling the
   * trunkward node directly, but then the creds would have to be sent to client.
   */
  public function accountNameFilter(string $acc_path = '', $limit = 10) : array {
    $result = [];
    try {
      $result = parent::accountNameFilter($acc_path, $limit);
    }
    catch (CCError $ex) {
      $this->logger->error(
        "accountNameFilter returned @type @message<br />@fullerror",
        ['@type' => get_class($ex), '@message' => $ex->makeMessage(), '@fullerror'=> $ex]
      );
    }
    catch (\Throwable $ex) {
      $this->logger->error(
        "accountNameFilter returned @type @message<br />@fullerror",
        ['@type' => get_class($ex), '@message' => $ex->getMessage(), '@fullerror'=> $ex]
      );
    }
    return $result;
  }

  /**
   * {@inheritDoc}
   * @todo upgrade the API
   */
  public function convertPrice(string $node_path) : \stdClass {

  }

  /**
   * {@inheritDoc}
   */
  public function buildValidateRelayTransaction($transaction) : array {
    try {
      $additional = parent::buildValidateRelayTransaction($transaction);
    }
    catch (\CreditCommons\Exceptions\TransactionLimitViolation $e) {
      $rate = \Drupal::config('mcapi_cc.settings')->get('trunkward.rate');
      // make a new exception that formats the value
      $args = [
        '@acc' => $e->acc_id,
        '@node' => $e->node,
        '@diff' => \Drupal\Mcapi\Functions::formatCurrency(round($e->diff / $rate), FALSE)
      ];
      // @todo how to render &#188; as 1/4 in the message box?
      if ($e instanceOf MaxLimitViolation) {
        $message = t("This transaction would put account '@acc' on node @node @diff above its maximum balance.", $args);
      }
      else {
        $message = t("This transaction would put account '@acc' on node @node @diff below its maximum balance.", $args);
      }
      throw new \Exception(Markup::create(strip_tags($message)));
    }
    // Convert remote addresses to local wallets.
    foreach($additional as &$entry) {
      $entry->payee = Functions::convertCCaddressToWalletId($entry->payee, $entry->metadata);
      $entry->payer = Functions::convertCCaddressToWalletId($entry->payer, $entry->metadata);
      $entry = (array)$entry;
      $entry['metadata'] = (array)$entry['metadata'];
    }
    return $additional;
  }

  public function transactionChangeState(string $uuid, string $target_state) : void {
    try {
      parent::transactionChangeState($uuid, $target_state);
    }
    catch(CCError $e) {
      throw new \Exception($e->getMessage());
    }
  }

  /**
   * Log every request to the network.
   * @param string $endpoint
   * @return \stdClass|NULL
   */
  protected function request(string $endpoint = '/') :\stdClass|NULL {
    $this->setHeader('CC-Auth', Functions::getLastHash());
    // Format the headers in an explodable array
    $headers = $this->options[RequestOptions::HEADERS] ? array_map( // cumbersome...
      function ($v, $k){return "$k: $v";},
      $this->options[RequestOptions::HEADERS],
      array_keys($this->options[RequestOptions::HEADERS])
    ) : [];
    try {
      $data = parent::request($endpoint);
      $this->logger->notice(
         "<p>Request @method @domain/@endpoint<br />@headers<br />@body</p><p>Response was:<pre>@response</pre></p>",
        [
          '@headers'  => Markup::create(implode('<br />', $headers)),
          '@method'   => strtoupper($this->method),
          '@domain'   => $this->baseUrl,
          '@endpoint' => $endpoint,
          '@body'     => Markup::Create($this->options[RequestOptions::BODY] ?? ""),
          '@response' => print_r($data, 1)
        ]
      );
    }
    catch(CCFailure $e) {
      // Logger doesn't cope with the long $e->trace, so format a friendler message.
      foreach ((array)$e as $prop => $val) {
        if ($prop == 'trace' or $prop == 'translated' or substr($prop, 1, 5) == 'Error')continue;
        $response_body[] = "$prop : ".print_r($val, 1);
      }
      $this->logger->error(
        "<p>Request @method @domain/@endpoint<br />@headers<br />@request_body</p><p>Response was @response_code: @response_message<br /><pre>@response_body</pre></p>",
        [
          '@headers'  => Markup::create(implode('<br />', $headers)),
          '@method'   => strtoupper($this->method),
          '@domain'   => $this->baseUrl,
          '@endpoint' => $endpoint,
          '@request_body'     => Markup::Create($this->options[RequestOptions::BODY] ?? ""),
          '@response_code'     => $e->getCode(),
          '@response_message'  => $e->makeMessage(),
          '@response_body' => implode('<br />', $response_body)
        ]
      );
      // Don't give any details to the user.
      // Drupal seems to be logging this exception, but the idea is to send a message to the user, but we don't have $form_state
      $this->messenger->addError('There was a problem on the remote server. See the log for more details.');
      $data = [];
    }
    return $data;
  }

}
