<?php

namespace Drupal\mcapi_forms;

use Drupal\mcapi\Functions;
use Drupal\mcapi\Entity\Transaction;
use Drupal\mcapi\Entity\Wallet;
use Drupal\system\Entity\Menu;
use Drupal\Core\Entity\EntityForm;
use Drupal\Core\Entity\Entity\EntityFormDisplay;
use Drupal\Core\Field\FieldItemList;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\Element\PathElement;
use Drupal\Core\Link;
use Drupal\Core\Url;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Extension\ModuleInstallerInterface;
use Drupal\Core\Cache\Cache;
use Drupal\Core\Menu\MenuParentFormSelectorInterface;
use Drupal\Core\Routing\RouteBuilderInterface;
use Drupal\Core\Routing\RequestContext;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Form to edit the FormEditor entity.
 */
class FormEditorForm extends EntityForm {

  const ENTRY_PROPERTIES = ['quant', 'payee', 'payee', 'description'];

  /**
   * @var ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * @var ModuleInstallerInterface
   */
  protected $moduleInstaller;

  /**
   * @var RouteBuilderInterface
   */
  protected $routerBuilder;

  /**
   * @var RequestContext
   */
  protected $routerRequestContext;

  /**
   * @var MenuParentFormSelectorInterface
   */
  protected $menuParentFormSelector;

  function __construct(
    ModuleHandlerInterface $module_handler,
    ModuleInstallerInterface $module_installer,
    RouteBuilderInterface $router_builder,
    RequestContext $router_request_context,
    MenuParentFormSelectorInterface $menu_parent_form_selector
  ) {
    $this->moduleHandler = $module_handler;
    $this->moduleInstaller = $module_installer;
    $this->routerBuilder = $router_builder;
    $this->routerRequestContext = $router_request_context;
    $this->menuParentFormSelector = $menu_parent_form_selector;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('module_handler'),
      $container->get('module_installer'),
      $container->get('router.builder'),
      $container->get('router.request_context'),
      $container->get('menu.parent_form_selector'),
    );
  }

  /**
   * {@inheritdoc}
   */
  public function form(array $form, FormStateInterface $form_state) {
    $form = parent::form($form, $form_state);
    $w = 0;
    $form['label'] = [
      '#title' => t('Label of the form'),
      '#type' => 'textfield',
      '#default_value' => $this->entity->label(),
      '#size' => 40,
      '#maxlength' => 80,
      '#required' => TRUE,
      '#weight' => $w++
    ];
    $form['id'] = [
      '#type' => 'machine_name',
      '#default_value' => $this->entity->id(),
      '#maxlength' => 12,
      '#machine_name' => [
        'exists' => '\Drupal\mcapi_forms\Entity\McForm::load',
      ],
      '#disabled' => !$this->entity->isNew(),
    ];
    $form['wallet_link_title'] = [
      '#title' => t('Link title to show on wallet'),
      '#description' => t('Link to this form from the wallet. Optionally use token [mc_wallet:name]'),
      '#type' => 'textfield',
      '#default_value' => $this->entity->getWalletLinkTitle(),
      '#weight' => $w++
    ];

    $form['workflow_type'] = [
      '#title' => t('Workflow type'),
      '#description' => $this->t('Users must have permission to use this workflow.'),
      '#type' => 'select',
      '#options' => Functions::entityOptions('mc_workflow'),
      '#default_value' => $this->entity->getWorkflowId(),
      '#required' => TRUE,
      '#weight' => $w++
    ];
    $form['path'] = [
      '#title' => t('Path'),
      '#description' => t('The url path at which this form will appear. Must begin with / and be unique.'),
      '#type' => 'textfield',
      '#default_value' => $this->entity->getPath(),
      '#placeholder' => '/path/to/form',
      '#pattern' => '\/[a-z\-0-9_/]+',
      '#convert_path' => PathElement::CONVERT_URL,
      '#element_validate' => [[get_class(), 'uniquePath']],
      '#required' => TRUE,
      '#field_prefix' => $this->routerRequestContext->getCompleteBaseUrl(),
      '#weight' => $w++
    ];
    $form['menu'] = [
      '#title' => t('Menu link'),
      '#type' => 'fieldset',
      '#tree' => TRUE,
      'title' => [
        '#title' => t('Menu link title'),
        '#description' => t('To make tabs, use the format link_title/tab_title.'),
        '#type' => 'textfield',
        '#default_value' => $this->entity->getMenuItem()->title ?? '',
        '#weight' => 1,
      ],
      'weight' => [
        '#title' => t('Weight'),
        '#type' => 'weight',
        '#default_value' => $this->entity->getMenuItem()->weight ?? 0,
        '#description' => t('In the menu, the heavier links will sink and the lighter links will be positioned nearer the top.'),
        '#weight' => 2,
      ],
      'add_transaction_tab' => [
        '#title' => t('Add tab to transactions page'),
        '#descriptions' => t('Tabs are ordered alphabetically'),
        '#type' => 'checkbox',
        '#default_value' => $this->entity->hasTransactionTab(),
        '#weight' => 3,
      ],
      '#weight' => $w++,
    ];

    if (!$this->moduleHandler->moduleExists('menu_ui')) {
      $this->moduleInstaller->install(['menu_ui']);
    }
    $custom_menus = [];
    foreach (Menu::loadMultiple() as $menu_parent => $menu) {
      $custom_menus[$menu_parent] = $menu->label();
    }
    $form['menu']['parent'] = $this->menuParentFormSelector
      ->parentSelectElement($settings['menu_parent'] ?? 'main:');
    $form['menu']['parent']['#empty_value'] = '';
    $form['menu']['parent']['#title'] = $this->t("Parent menu item");
    $form['menu']['parent']['#default_value'] = $this->entity->getMenuItem()->parent??'';
    $form['presets'] = [
      '#type' => 'details',
      //'#open' => !empty(array_filter($this->entity->presets)), //todo open if anything is set
      '#tree' => TRUE,
      '#title' => $this->t('Preset values'),
      '#description' => $this->t('Fields with preset values are not required in the twig template'),
      '#weight' => $w++,
      'description' => [
        '#title' => $this->t('Description'),
        '#type' => 'textfield',
        '#default_value' => $this->entity->getPreset('description'),
      ],
      'quant' => [
        '#title' => $this->t('Quantity'),
        '#type' => 'worth',
        '#default_value' => $this->entity->getPreset('quant'),
      ],
      'payee_wid' => [
        '#title' => $this->t('Payee wallet'),
        '#type' => 'wallet_entity_auto',
        '#default_value' => $this->entity->getPreset('payee_wid') ? Wallet::load($this->entity->getPreset('payee_wid')) : NULL,
      ],
      'payer_wid' => [
        '#title' => $this->t('Payer wallet'),
        '#type' => 'wallet_entity_auto',
        '#default_value' => $this->entity->getPreset('payer_wid') ? Wallet::load($this->entity->getPreset('payer_wid')) : NULL,
      ]
    ];

    $form_display = EntityFormDisplay::load('mc_transaction.mc_transaction.default');
    foreach ($this->getAdditionalFieldDefinitions() as $fname => $field_definition) {
      $items = $this->dummyFieldItems($fname, $this->entity->getPreset($fname));
      $field_def = clone($field_definition);
      $field_def->setRequired(FALSE);
      $widget = \Drupal::service('plugin.manager.field.widget')->getInstance([
        'field_definition' => $field_def,
        'form_mode' => 'default',
        // No need to prepare, defaults have been merged in setComponent().
        'prepare' => TRUE,
        'configuration' => $form_display->getComponent($fname),
      ]);
      $element = ['#parents' => ['presets']];
      $form['presets'][$fname] = $widget->form($items, $element, $form_state)['widget'];
      $form['presets'][$fname]['#weight'] = 20;
    }
    $form['twig'] = [
      '#title' => t('Main form @Twig', ['@Twig' => 'Twig']),
      '#description' => t(
        'Use the following @Twig tokens with HTML & css to design your payment form. Linebreaks will be replaced automatically.',
        ['@Twig' => Link::fromTextAndUrl('Twig', Url::fromUri('http://twig.sensiolabs.org/doc/templates.html'))->toString()]
      ),
      '#type' => 'textarea',
      '#rows' => 6,
      '#default_value' => str_replace(array('\r\n', '\n', '\r'), "\n", $this->entity->getTwigTemplate()),
      '#element_validate' => [[$this, 'validateTwigTemplate']],
      '#required' => TRUE,
      '#weight' => $w++
    ];

    $required_properties = array_combine(self::ENTRY_PROPERTIES, self::ENTRY_PROPERTIES);
    if ($workflow = $this->entity->getWorkflow()) {
      if ($workflow->direction <> 'bill') {
        unset($form['presets']['payer']);
      }
      if ($workflow->direction <> 'credit') {
        unset($form['presets']['payee']);
      }
      if ($workflow->direction == 'bill') {
        unset($required_properties['payer']);
      }
      elseif ($workflow->direction == 'credit') {
        unset($required_properties['payee']);
      }
    }
    list($required_additional_fields, $optional_tokens) = $this->getRequiredFieldAPI();

    $form['twig']['#description'] .= t(
      'These fields must be included AND/OR preset: @fields.',
      ['@fields' => ' {{ ' . implode(' }}, {{ ', array_merge($required_additional_fields, $required_properties)) . ' }}']
    );
    if ($optional_tokens) {
      $form['twig']['#description'] .= ' '.t(
        'These fields are optional: @fields.',
        ['@fields' => ' {{ ' . implode(' }}, {{ ', $optional_tokens) . ' }}']
      );
    }

    $form['quicklink'] = [
      '#title' => t('Show quicklink on wallet'),
      '#descriptions' => t('Only if the wallet extrafield '),
      '#type' => 'checkbox',
      '#default_value' => $this->entity->getQuicklink(),
      '#weight' => $w++
    ];

    return $form;
  }

  /**
   * {@inheritDoc}
   */
  private function getTransactionWidget($field_name) {
    return EntityFormDisplay::load('mc_transaction.mc_transaction.default')
      ->getRenderer($field_name);
  }

  /**
   * {@inheritDoc}
   */
  public function save(array $form, FormStateInterface $form_state) {
    $status = $this->entity->save();
    $this->routerBuilder->rebuild();
    Cache::invalidateTags(['route_info']);

    if ($status == SAVED_UPDATED) {
      $this->messenger()->addStatus(t('Form %label has been updated.', ['%label' => $this->entity->label()]));
    }
    else {
      $this->messenger()->addStatus(t('Form %label has been added.', ['%label' => $this->entity->label()]));
    }
    $form_state->setRedirect('entity.entity_form_display.mc_transaction.default');
  }

  /**
   * Element validation callback ensuring that the given path is unique.
   */
  public static function uniquePath(&$element, $form_state) {
    $path = $form_state->getValue('path');
    $url = Url::fromUserInput($path);
    if ($url->isRouted()) {
      $params = $url->getRouteParameters();
      if (isset($params['mc_workflow']) and $params['mc_workflow'] <> $form_state->getFormObject()->entity->id()) {
        $form_state->setError($element, t('The path %path already exists.', ['%path' =>  $path]));
      }
    }
  }


  /**
   * Check the required transaction form elements are either preset OR in Twig.
   */
  public function validateTwigTemplate(array &$element, $form_state) {
    $twig_template = $element['#value'];
    preg_match_all('/\{\{ ?([a-z_]+) ?\}\}/', $twig_template, $matches);
    $used_tokens = $matches[1] ?? [];
    // Check for duplicate tokens.
    if (count($used_tokens) <> count(array_unique($used_tokens))) {
      $form_state->setError($element, t('Each twig token can only appear once in the form.'));
    }
    // what's in the twig will incorporate what is in the presets
    foreach ($form_state->getValue('presets') as $k => $v) {
      if (!empty($v)) {
        if (substr($k, -4) == '_wid') $k = substr($k, 0, -4);
        $used_tokens[] = $k;
      }
    }

    list($required_api_fields) = $form_state->getFormObject()->getRequiredFieldAPI();
    $preset_values = array_filter($form_state->getValue('presets'));
    // Check for missing required tokens.
    $additional_fields = array_diff($required_api_fields, array_keys($preset_values), Self::ENTRY_PROPERTIES);
    foreach ($additional_fields as $fname) {
      if (!in_array($fname, $used_tokens)) {
        $field_items = $this->dummyFieldItems($fname, $form_state->getvalue('presets.fname'));
        if ($field_items->isEmpty()) {
          $form_state->setError($element, t('Missing required token @token', ['@token' => '{{ '.$fname.' }}']));
        }
      }
    }
  }

  /**
   * Make a dummy transaction fieldItemList with a value already set.
   */
  private function dummyFieldItems($field_name, $field_value) : FieldItemList {
    return Transaction::create(['workflow' => 'default_bill', $field_name => $field_value])->get($field_name);
  }

  /**
   * Get the transaction entity's required and optional form fields
   * @return array
   */
  public function getRequiredFieldAPI() : array {
    // Fields that cannot be set
    $required_field_api = $optional_field_api = [];
    $fixed = ['xid', 'uuid', 'serial', 'creator', 'workflow', 'state', 'changed', 'created', 'entries', 'vxid', 'signatures', 'revision_default'];
    foreach ($this->getAdditionalFieldDefinitions() as $id => $field_def) {
      /** @var \Drupal\Core\Field\BaseFieldDefinition $field_def */
      if (in_array($id, $fixed)) {
        continue;
      }
      if ($field_def->isRequired()) {
        $required_field_api[] = $id;
      }
      else {
        $optional_field_api[] = $id;
      }
    }
    return [$required_field_api, $optional_field_api];
  }

  /**
   * Convenience
   */
  private function getAdditionalFieldDefinitions() :array {
    $entity_field_manager = \Drupal::service('entity_field.manager');
    $all_fields = $entity_field_manager->getFieldDefinitions('mc_transaction', 'mc_transaction');
    $base_fields = $entity_field_manager->getBasefieldDefinitions('mc_transaction');
    return array_diff_key($all_fields, $base_fields);
  }

}
