<?php

namespace Drupal\mcapi\Entity\Storage;

use Drupal\mcapi\Entity\Transaction;
use Drupal\mcapi\Entity\TransactionInterface;
use Drupal\Core\Entity\Sql\SqlContentEntityStorage;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Database\Query\Query;
use Drupal\Core\Database\Query\Select;
use Drupal\Core\Cache\Cache;

/**
 * Base transaction storage controller for the transaction index tables.
 *
 * Stores everything in a table which is designed to work with views.
 */
abstract class TransactionIndexStorage extends SqlContentEntityStorage implements TransactionStorageInterface {

  /**
   * {@inheritDoc}
   *
   * Save 2 rows per entry into the index tables.
   */
  public function doSave($id, EntityInterface $entity) {
    $return = parent::doSave($id, $entity);
    $this->database->delete('mc_transactions_index')
      ->condition('xid',$entity->id())
      ->execute();
    if (in_array ($entity->state->value, TransactionInterface::COUNTED_STATES)) {// Determines if the current state is counted.
      $record = $this->mapToStorageRecord($entity);
      $common = [
        'xid' => $record->xid,
        'state' => $record->state,
        'workflow' => $record->workflow,
        'created' => $record->created// $entity->get('created')->value
      ];
      foreach ($entity->entries->getValue() as $delta => $entry) {
        $this->database->insert('mc_transactions_index')
          ->fields($common + [
            'delta' => $delta,
            'wallet_id' => $entry['payer_wid'],
            'partner_id' => $entry['payee_wid'],
            'incoming' => 0,
            'outgoing' => $entry['quant'],
            'diff' => -$entry['quant'],
            'volume' => $entry['quant'],
            'primary' => (int)$entry['primary']
          ])
          ->execute();

        $this->database->insert('mc_transactions_index')
          ->fields($common + [
            'xid' => $record->xid,
            'delta' => $delta,
            'wallet_id' => $entry['payee_wid'],
            'partner_id' => $entry['payer_wid'],
            'incoming' => $entry['quant'],
            'outgoing' => 0,
            'diff' => $entry['quant'],
            'volume' => $entry['quant'],
            'primary' => (int)$entry['primary']
          ])
          ->execute();
      }
      // Now this is saved, we update the transaction totals table from it.
      $this->recalcTransactionTotals($entity->allWalletIds());
    }
    return $return;
  }

  /**
   * {@inheritDoc}
   */
  protected function doDelete($entities) {
    parent::doDelete($entities);
    $this->database
      ->delete('mc_transactions_index')
      ->condition('xid', array_keys($entities), 'IN')
      ->execute();
  }

    private function invalidateWalletCache(Transaction $transaction) {
    $tags = [];
    foreach (array_chunk($transaction->allWalletIds(), 50) as $wids) {
      foreach (Wallet::loadMultiple($wids) as $wallet) {
        $tags = array_merge($tags, $wallet->getCacheTagsToInvalidate());
      }
    }
    Cache::invalidateTags($tags);
  }

  /**
   * {@inheritDoc}
   * reset the cache of all the wallets involved as well.
   */
  public function resetCache(array $xids = NULL) {
    parent::resetCache($xids);
    $tags = ['mc_wallet_values'];
    if ($xids) {
      $wids = $cids = [];
      foreach(Transaction::loadMultiple($xids) as $transaction) {
        $wids = array_merge($wids, $transaction->allWalletIds());
      }
      $wids = array_unique($wids);
      $this->recalcTransactionTotals($wids);

      $tags = [];
      $wallet_storage = \Drupal::entitytypeManager()->getStorage('mc_wallet');

      foreach (array_chunk($wids, 50) as $wids_chunk) {
        foreach ($wallet_storage->loadMultiple($wids_chunk) as $wallet) {
          $tags = array_merge($tags, $wallet->getCacheTagsToInvalidate());
          $cids[] = $wallet_storage->buildCacheId($wallet->id());
        }
      }
      $this->cacheBackend->deleteMultiple($cids);
    }
    Cache::invalidateTags($tags);
  }

  /**
   * {@inheritDoc}
   */
  public function indexRebuild() : void {
    $this->database->truncate('mc_transactions_index')->execute();
    // don't know how to do this with database API.
    $this->database->query("
      INSERT INTO {mc_transactions_index} (
        SELECT
          t.[xid],
          e.[delta],
          e.[entries_payer_wid] AS wallet_id,
          e.[entries_payee_wid] AS partner_id,
          t.[state],
          t.[workflow],
          t.[created],
          0 AS incoming,
          e.[entries_quant] AS outgoing,
          - e.[entries_quant] AS diff,
          e.[entries_quant] AS volume,
          e.[entries_primary] AS `primary`
        FROM {mc_transaction} t
        RIGHT JOIN {mc_transaction__entries} e ON t.[xid] = e.[entity_id]
        WHERE [state] IN ('completed')
      )"
    );
    $this->database->query(
      "INSERT INTO {mc_transactions_index} (
        SELECT
          t.[xid],
          e.[delta],
          e.[entries_payee_wid] AS wallet_id,
          e.[entries_payer_wid] AS partner_id,
          t.[state],
          t.[workflow],
          t.[created],
          e.[entries_quant] AS incoming,
          0 AS outgoing,
          e.[entries_quant] AS diff,
          e.[entries_quant] AS volume,
          e.[entries_primary] AS `primary`
        FROM {mc_transaction} t
        RIGHT JOIN {mc_transaction__entries} e ON t.[xid] = e.[entity_id]
        WHERE [state] IN ('completed')
      ) "
    );

    $this->database->truncate('mc_transaction_totals')->execute();
    $result = $this->database->select('mc_transactions_index', 'i')
      ->fields('i', ['wallet_id'])
      ->distinct()
      ->execute();
    $wids = [];
    foreach ($result->fetchAll() as $bal) {
      $wids[] = $bal->wallet_id;
    }
    $this->recalcTransactionTotals($wids);
  }

  /**
   * {@inheritDoc}
   */
  public function indexCheck() : bool {
    $q = $this->database->select('mc_transactions_index');
    $q->addExpression('SUM(diff)');
    // If the sum of everything is zero.
    if ($q->execute()->fetchfield() + 0 != 0) {
      \Drupal::messenger()->addError('Transaction index table rows do not add up to zero');
      return FALSE;
    }
    $q = $this->database->select('mc_transactions_index', 'i');
    $q->addExpression('SUM(incoming)');
    $q->condition('i.state', TransactionInterface::COUNTED_STATES, 'IN');
    $volume_index = $q->execute()->fetchField();

    $q = $this->database->select('mc_transaction', 't');
    $q->join('mc_transaction__entries', 'e', 't.xid = e.entity_id');
    $q->addExpression('SUM(e.entries_quant)');
    $q->condition('t.state', TransactionInterface::COUNTED_STATES, 'IN');
    $volume = $q->execute()->fetchField();
    if ($volume_index == $volume) {
      return TRUE;
    }
    \Drupal::messenger()->addError('Transaction volumes in main table and index table do not correspond');
    return FALSE;
  }

  /**
   * {@inheritDoc}
   */
  public function historyOfWallet($wid, $since = 0) : array {
    // This is a way to add up the results as we go along.
    $this->database->query("SET @csum := 0");
    // Have to calculate the whole history by adding up all the diffs
    // It is cheaper to do stuff in mysql.
    $query = $this->database->select('mc_transactions_index', 'i')->fields('i', ['created']);
    $query->addExpression('(@csum := @csum + [diff])', '[balance]');
    $all_balances = $query->condition('wallet_id', $wid)
      ->orderby('xid', 'ASC')
      ->execute()
      // If two transactions happen on the same second, the latter running
      // balance will be shown only.
      ->fetchAllKeyed();
    $all_balances = [(int)key($all_balances)-1 => 0] + $all_balances;
    $pos = 0;
    // Having done the addition, we can chop the beginning off the array.
//    if ($since) {
//      // We know they keys are in chronological order.
//      foreach (array_keys($all_balances) as $pos => $created) {
//        if ($created > $since) {
//          // Now the $pos is set.
//          break;
//        }
//      }
//      // Cut the beginning off
//      // to the end.
//      $all_balances = array_slice($all_balances, $pos);
//    }
    $all_balances[\Drupal::time()->getRequestTime()] = end($all_balances);
    return $all_balances;
  }

  /**
   * Add an array of conditions to the select query.
   *
   * This is more flexible than entityQuery which runs on the transaction table
   * itself. This has no access control.
   *
   * @param array $conditions
   *   Conditions keyed by property name suitable for getmcapiIndexQuery():
   *   xid, partner_id, wallet_id, workflow, state, involving, since, until, value,
   *   min, max
   *
   * @note This query may return double the expected results, unless filtered appropriately
   */
  private function getMcapiIndexQuery(array $conditions = []) : Select {
    $query = $this->database
      ->select('mc_transactions_index', 'i');
    if (empty($conditions['state'])) {
      $conditions['state'] = TransactionInterface::COUNTED_STATES;
    }
    foreach ($conditions as $field => $value) {
      if ($value === NULL) {
        //\Drupal::logger('mcapi')->debug("getmcapiIndexquery $field needs debugging when value = NULL");
        continue;
      }
      switch ($field) {
        // Case 'payer': these can't work
        // case 'payee':
        case 'xid':
        case 'wallet_id':
        case 'partner_id':
        case 'workflow':
        case 'state':
          $query->condition($field, (array) $value, 'IN');
          break;

        case 'involving':
          $value = (array) $value;
          $cond_group = count($value) == 1 ? $this->database->or() : $this->database->and();
          $query->condition($cond_group
            ->condition('wallet_id', $value)
            ->condition('partner_id', $value)
          );
          break;

        case 'since':
          $query->condition('created', $value, '>');
          break;

        case 'until':
          $query->condition('created', $value, '<');
          break;

        // Synonyms as far as the index table is concerned.
        case 'value':
          $query->condition('volume', $value);
          break;

        // Synonyms as far as the index table is concerned.
        case 'min':
          $query->condition('volume', $value, '>=');
          break;

        // Synonyms as far as the index table is concerned.
        case 'max':
          $query->condition('volume', $value, '<=');
          break;

        default:
          \Drupal::logger('mcapi')->error('Filtering on unknown field: @fieldname', ['@fideldname' => $field]);
      }
    }
    return $query;
  }

  /**
   * {@inheritDoc}
   */
  public function wallets($conditions = []) : array {
    return $this->getmcapiIndexQuery($conditions)
      ->fields('i', ['wallet_id'])
      ->distinct()
      ->execute()
      ->fetchCol();
  }

  /**
   * {@inheritDoc}
   *
   * @note The running balance views field isn't used in any default views.
   */
  public function runningBalance($wid, $field_name = 'xid', $before = 0) : int {
    // The running balance depends the order of the transactions. we will assume
    // the order of creation is what's wanted because that corresponds to the
    // order of the xid. NB it is possible to change the apparent creation date.
    $query = $this->database->select('mc_transactions_index');
    $query->addExpression('SUM([diff])');
    $query->condition('wallet_id', $wid);
    if ($before) {
      $query->condition($field_name, $before, '<=');
    }
    return $query->execute()->fetchField();
  }

  /**
   * {@inheritDoc}
   */
  public function balanceByWorkflow(int $wid, $workflow_ids, $states = []) : int {
    $args = ['wallet_id' => $wid, 'workflow' => $workflow_ids];
    if ($states) {
      $args['state'] = $states;
    }
    $q = $this->getmcapiIndexQuery($args);
    $q->addExpression('SUM([diff])', 'balance');
    return $q->execute()->fetchField() ?? 0;
  }

  /**
   * {@inheritDoc}
   *
   * @note If there are no transactions balance, volume, income, expenditure will be NULL
   */
  public function ledgerStateQuery(array $conditions) : Query {
    $q = $this->getmcapiIndexQuery($conditions);
    $q->addExpression('SUM([diff])', 'balance');
    $q->addExpression('SUM([volume])/2', 'volume');
    $q->addExpression('SUM([incoming])', 'income');
    $q->addExpression('SUM([outgoing])', 'expenditure');
    $q->addExpression('COUNT([partner_id])', 'trades');
    $q->addExpression('COUNT (DISTINCT [partner_id])', 'partners');
    return $q;
  }

  /**
   * {@inheritDoc}
   */
  public function ledgerStateByWallet(array $conditions) : array {
    $q = $this->ledgerStateQuery($conditions);
    $q->addExpression('wallet_id', 'wid');
    $q->groupby('wallet_id');
    return $q->execute()->fetchAll();
  }

  /**
   * {@inheritDoc}
   */
  public function historyPeriodic($period, $conditions) : array {
    $q = $this->getmcapiIndexQuery($conditions);
    $q->addExpression('COUNT(DISTINCT [xid])', 'trades');
    $q->addExpression('SUM([incoming])', 'volumes');
    $q->addExpression('COUNT(DISTINCT [wallet_id])', 'wallets');
    // Unfortunately all these numbers are retrieved as strings.
    $q->condition('incoming', 0, '>=');
    switch ($period) {
      case 'day':
        $q->addExpression('DATE(FROM_UNIXTIME([created]))', 'date');
        $q->groupBy('date');
        break;

      case 'week':
        $q->addExpression('WEEK(FROM_UNIXTIME([created]))', 'weeknum');
        $q->addExpression('MIN(YEAR(FROM_UNIXTIME([created])))', 'year');
        $q->groupBy('weeknum');
        break;

      case 'month':
        $q->addExpression('MONTH(FROM_UNIXTIME([created]))', 'month');
        $q->addExpression('MIN(YEAR(FROM_UNIXTIME([created])))', 'year');
        $q->groupBy('month');
        break;

      case 'year':
        $q->addExpression('YEAR(FROM_UNIXTIME([created]))', 'year');
        $q->groupBy('year');
        break;
    }
    $dates = $trades = $volumes = $wallets = [];
    $thisyear = date('Y');
    foreach ($q->execute()->fetchAll() as $row) {
      switch ($period) {
        case 'day':
          $dates[] = strtotime($row->date);
          break;
        case 'week':
          // The last second of the week.
          $dates[] = strtotime($row->year . 'W' . str_pad($row->weeknum + 1, 2, 0, STR_PAD_LEFT)) - 1;
          break;
        case 'month':
          // The last second of the month using the format YYYY-MM
          $dates[] = strtotime($row->year . '-' . ($row->month + 1)) - 1;
          break;
        case 'year':
          if ($row->year == $thisyear) {
            $dates[] = \Drupal::service('datetime.time')->getCurrentTime();
          }
          else {
            $dates[] = strtotime('01-01-'.$row->year)-1;
          }
      }
      $trades[] = intval($row->trades);
      $volumes[] = intval($row->volumes);
      $wallets[] = intval($row->wallets);
    }
    return [$dates, $trades, $volumes, $wallets];
  }

  /**
   * Update the transaction totals table.
   *
   * @param array $wids
   */
  private function recalcTransactionTotals(array $wids) {
    foreach ($wids as $wid) {
      $q1 = $this->getmcapiIndexQuery(['wallet_id' => $wid]);
      $q1->addExpression('COUNT(DISTINCT i.[xid])', 'trades');
      $q1->addExpression('COALESCE(SUM(i.[incoming]), 0)', 'gross_in');
      $q1->addExpression('COALESCE(SUM(i.[outgoing]), 0)', 'gross_out');
      $q1->addExpression('COALESCE(SUM(i.[diff]), 0)', 'balance');
      $q1->addExpression('COALESCE(SUM(i.[volume]), 0)', 'volume');
      $q1->addExpression('COUNT(DISTINCT i.[partner_id])', 'partners');
      $fields = $q1->execute()->fetch(\PDO::FETCH_ASSOC);
      $this->database->merge('mc_transaction_totals')
        ->keys(['wid' => $wid])
        ->fields($fields)
        ->execute();
    }

  }

}
