<?php

namespace Drupal\mcapi\Entity\ViewBuilder;

use Drupal\mcapi\Entity\Transaction;
use Drupal\Core\Render\Markup;
use Drupal\Core\Config\ConfigFactory;
use Drupal\Core\Routing\CurrentRouteMatch;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\EntityViewBuilder;
use Symfony\Component\DependencyInjection\ContainerInterface;
use CreditCommons\TransactionInterface as CCTransactionInterface;

/**
 * Render controller for transactions.
 */
class TransactionViewBuilder extends EntityViewBuilder {

  protected $settings;
  protected $routeMatch;

  /**
   *
   * @param type $entity_type
   * @param type $entity_repository
   * @param type $language_manager
   * @param type $theme_registry
   * @param type $entity_display_repository
   * @param ConfigFactory $config_factory
   * @param CurrentRouteMatch $route_match
   * @param ModuleHandler $module_handler
   */
  public function __construct($entity_type, $entity_repository, $language_manager, $theme_registry, $entity_display_repository, ConfigFactory $config_factory, CurrentRouteMatch $route_match) {
    parent::__construct($entity_type, $entity_repository, $language_manager, $theme_registry, $entity_display_repository);
    $this->settings = $config_factory->get('mcapi.settings');
    $this->routeMatch = $route_match;
  }

  /**
   * {@inheritDoc}
   *
   * @todo update with entity_type.manager when core interface changes
   */
  public static function createInstance(ContainerInterface $container, EntityTypeInterface $entity_type) {
    return new static(
      $entity_type,
      $container->get('entity.repository'),
      $container->get('language_manager'),
      $container->get('theme.registry'),
      $container->get('entity_display.repository'),
      $container->get('config.factory'),
      $container->get('current_route_match')
    );
  }

  /**
   * {@inheritDoc}
   */
  protected function getBuildDefaults(EntityInterface $entity, $view_mode) {
    switch ($view_mode) {
      case 'full':
        // Prepare the rendered fields from the ViewDisplay
        $build = parent::getBuildDefaults($entity, $view_mode);
        $build['#links'] = [];
        if ($this->routeMatch->getRouteName() == 'entity.mc_transaction.canonical') {
          $build['#links'] = $this->linkList($entity);
        }
        switch ($entity->state->value) {
          case CCTransactionInterface::STATE_PENDING:
            $build['#watermark'] = t('Pending', [], ['context' => 'Transaction certificate watermark']);
            break;
          case CCTransactionInterface::STATE_ERASED:
            $build['#watermark'] = t('Void', [], ['context' => 'Transaction certificate watermark']);
        }
        $build['#attributes']['class'][] = 'certificate';
        break;

      case 'sentence':
        // Prepare using the token system
        $template = $this->settings->get('sentence_template') . ' [xaction:links]';
        $sentence =  \Drupal::token()->replace($template, ['xaction' => $entity]);
        $linklist = $this->linkList($entity);
        $links = \Drupal::service('renderer')->render($linkList);

        $build = [
          //'#prefix' => '<div class = "mc-transaction-sentence">',
          '#markup' => Markup::create($sentence . $links),
          //'#suffix' => '</div>'
        ];
        $build['#attributes']['class'][] = 'sentence';
        unset($build['#theme']);
        break;

      default:
        throw new \Exception("Unrecognised transaction view mode: $view_mode");
    }
    $build['#attributes']['class'][] = 'mc-transaction';
    $build['#attributes']['class'][] = 'type-' . $entity->workflow->target_id;
    $build['#attributes']['class'][] = 'state-' . $entity->state->value;
    $build['#attached']['library'][] = 'mcapi/transaction';

    // @todo we might need to use the post-render cache to get the links right instead of template_preprocess_mc_transaction
    return $build;
  }

  /**
   * Make a renderable array for transaction operations
   *
   * @param Transaction $transaction
   *
   * @return array
   *   A renderable array.
   */
  public static function linkList(Transaction $transaction) : array {
    $links = \Drupal::entityTypeManager()
      ->getListBuilder('mc_transaction')
      ->getOperations($transaction);
    // $the links are in order
    return ['#theme' => 'links', '#links' => $links];
  }

}
