<?php

namespace Drupal\Mcapi;

use Drupal\mcapi\Entity\Transaction;
use Drupal\mcapi\Entity\TransactionInterface;
use Drupal\Core\Link;
use Drupal\Core\Render\Markup;
use Drupal\Component\Utility\Html;
use CreditCommons\TransactionInterface as CCTransactionInterface;

/**
 * Utility class for community accounting.
 */
class Functions {

  /**
   * Get a list of entity labels keyed by entity ID suitable for a select box.
   * @param string $entity_type_id
   * @return string[]
   *
   * @note this is only used for mc_state, mc_workflow ATM
   */
  static function entityOptions(string $entity_type_id) : array {
    $entities = \Drupal::entityTypeManager()->getStorage($entity_type_id)->loadMultiple();
    $contextRepo = \Drupal::service('entity.repository');
    foreach ($entities as $entity) {
      $label = $contextRepo->getTranslationFromContext($entity)->label();
      $list[$entity->id()] = Html::escape($label);
    }
    return $list;
  }

  /**
   * Get a list of all the tokens on the transaction entity.
   *
   * @note This really needs replacing with a core function.
   */
  public static function tokenHelp() : string {
    foreach (array_keys(\Drupal::Token()->getInfo()['tokens']['xaction']) as $token) {
      $tokens[] = "[xaction:$token]";
    }
    return implode(', ', $tokens);
  }

  /**
   * Format a raw currency value and add a link to the stats page.
   *
   * @staticvar type $formatter
   * @param int $raw_num
   * @param type $linked
   *   Link it to the metrics page.
   * @return Markup
   */
  public static function formatCurrency(int $raw_num, $linked = TRUE) : Markup {
    static $formatter;
    if (!isset($formatter)) {
      $formatter = \Drupal::Service('mcapi.currency_formatter');
    }
    // Replace the parts back into the number format.
    $output = $formatter->format($raw_num);
    // Link it to the currency stats page.
    if ($linked) {
      $output = (string)Link::createFromRoute(['#markup' => $output], 'mcapi.metrics')->toString();
    }
    return Markup::create($output);
  }

  /**
   * Get the transactions needing signature of the given user.
   *
   * @param int $uid
   *   The user ID.
   *
   * @return int[]
   *   Transaction IDs
   */
  static function transactionsNeedingSigOfUser($uid) : array {
    return \Drupal::entityQuery('mc_transaction')
      ->condition('signatures.uid', $uid)
      ->condition('signatures.signed', 0)
      ->accessCheck(FALSE)
      ->execute();
  }

  /**
   * Transactions which the user has signed but are still pending.
   * @param int $uid
   * @return int[]
   *   Transaction IDs
   */
  static function transactionsNeedingSig(int $uid) : array {
    return \Drupal::entityQuery('mc_transaction')
      ->condition('signatures.uid', $uid)
      ->condition('state', CCTransactionInterface::STATE_PENDING)
      ->accessCheck(FALSE)
      ->execute();
  }

  /**
   * Assembles transaction and returns the transaction with added entries in its
   * creation->state. To ensure the transaction is built only once, this stores
   * it as a static. This means it can be used to alter a passed transaction or
   * returns a previously passed transaction.
   *
   * @staticvar type $assembled
   * @param TransactionInterface $transaction
   * @return TransactionInterface
   *
   * @note this method cannot belong in the transaction object.
   */
  static function getAssembledTransaction(TransactionInterface $transaction = NULL) : TransactionInterface {
    static $assembled = [];
    $uuid = $transaction->uuid();
    if (!$transaction->isNew()) {
       throw new \Exception('Called Functions::getAssembledTransaction() with a non-new transaction.');
    }
    if (isset($assembled[$uuid])) {
      $transaction = $assembled[$uuid];
    }
    else {
      // Set the transaction to its initial state and add signatories.
      // Allow other modules to add entries, but not to modify the transaction.
      $entries = \Drupal::moduleHandler()->invokeAll('transaction_assemble', [clone($transaction)]);
      foreach ($entries as $params) {
        $params['primary'] = FALSE;
        $transaction->addEntry($params);
      }
      $assembled[$uuid] = $transaction;
    }
    return $transaction;
  }

  /**
   * Utility for quick rendering of multiple transactions as single sentences.
   *
   * @param int[] $xids
   * @return []
   *   A renderable array
   */
  static function renderSentences(array $xids) : array {
    $transactionViewBuilder = \Drupal::entityTypeManager()->getViewBuilder('mc_transaction');
    foreach (Transaction::loadMultiple($xids) as $t) {
      $renderable[] = $transactionViewBuilder->view($t, 'sentence');
    }
    return $renderable;
  }

}
